package java.awt.font;

import sun.font.BidiUtils;
import sun.font.TextLabelFactory;
import java.util.Map;
import java.awt.Font;
import java.util.Hashtable;
import sun.font.AttributeValues;
import java.text.AttributedCharacterIterator;
import java.text.BreakIterator;
import sun.font.TextLineComponent;
import java.text.Bidi;

public final class TextMeasurer implements Cloneable
{
    private static float EST_LINES;
    private FontRenderContext fFrc;
    private int fStart;
    private char[] fChars;
    private Bidi fBidi;
    private byte[] fLevels;
    private TextLineComponent[] fComponents;
    private int fComponentStart;
    private int fComponentLimit;
    private boolean haveLayoutWindow;
    private BreakIterator fLineBreak;
    private CharArrayIterator charIter;
    int layoutCount;
    int layoutCharCount;
    private StyledParagraph fParagraph;
    private boolean fIsDirectionLTR;
    private byte fBaseline;
    private float[] fBaselineOffsets;
    private float fJustifyRatio;
    private int formattedChars;
    private static boolean wantStats;
    private boolean collectStats;
    
    public TextMeasurer(AttributedCharacterIterator attributedCharacterIterator, FontRenderContext fFrc) {
        this.fLineBreak = null;
        this.charIter = null;
        this.layoutCount = 0;
        this.layoutCharCount = 0;
        this.fJustifyRatio = 1.0f;
        this.formattedChars = 0;
        this.collectStats = false;
        this.fFrc = fFrc;
        this.initAll(attributedCharacterIterator);
    }
    
    @Override
    protected Object clone() {
        TextMeasurer textMeasurer;
        try {
            textMeasurer = (TextMeasurer)super.clone();
        }
        catch (CloneNotSupportedException ex) {
            throw new Error();
        }
        if (this.fComponents != null) {
            textMeasurer.fComponents = (TextLineComponent[])this.fComponents.clone();
            return textMeasurer;
        }
        return textMeasurer;
    }
    
    private void invalidateComponents() {
        int length = this.fChars.length;
        this.fComponentLimit = length;
        this.fComponentStart = length;
        this.fComponents = null;
        this.haveLayoutWindow = false;
    }
    
    private void initAll(AttributedCharacterIterator paragraph) {
        this.fStart = paragraph.getBeginIndex();
        this.fChars = new char[paragraph.getEndIndex() - this.fStart];
        int n = 0;
        for (int i = paragraph.first(); i != '\uffff'; i = paragraph.next()) {
            this.fChars[n++] = i;
        }
        paragraph.first();
        this.fBidi = new Bidi(paragraph);
        if (this.fBidi.isLeftToRight()) {
            this.fBidi = null;
        }
        paragraph.first();
        Map<AttributedCharacterIterator.Attribute, Object> attributes = paragraph.getAttributes();
        NumericShaper numericShaping = AttributeValues.getNumericShaping(attributes);
        if (numericShaping != null) {
            numericShaping.shape(this.fChars, 0, this.fChars.length);
        }
        this.fParagraph = new StyledParagraph(paragraph, this.fChars);
        this.fJustifyRatio = AttributeValues.getJustification(attributes);
        if (TextLine.advanceToFirstFont(paragraph)) {
            Font fontAtCurrentPos = TextLine.getFontAtCurrentPos(paragraph);
            int beginIndex = paragraph.getIndex() - paragraph.getBeginIndex();
            LineMetrics lineMetrics = fontAtCurrentPos.getLineMetrics(this.fChars, beginIndex, beginIndex + 1, this.fFrc);
            this.fBaseline = (byte)lineMetrics.getBaselineIndex();
            this.fBaselineOffsets = lineMetrics.getBaselineOffsets();
        }
        else {
            this.fBaseline = TextLayout.getBaselineFromGraphic((GraphicAttribute)attributes.get(TextAttribute.CHAR_REPLACEMENT));
            this.fBaselineOffsets = new Font(new Hashtable(5, 0.9f)).getLineMetrics(" ", 0, 1, this.fFrc).getBaselineOffsets();
        }
        this.fBaselineOffsets = TextLine.getNormalizedOffsets(this.fBaselineOffsets, this.fBaseline);
        this.invalidateComponents();
    }
    
    private void generateComponents(int n, int n2) {
        if (this.collectStats) {
            this.formattedChars += n2 - n;
        }
        TextLabelFactory textLabelFactory = new TextLabelFactory(this.fFrc, this.fChars, this.fBidi, 0);
        int[] inverseMap = null;
        if (this.fBidi != null) {
            this.fLevels = BidiUtils.getLevels(this.fBidi);
            inverseMap = BidiUtils.createInverseMap(BidiUtils.createVisualToLogicalMap(this.fLevels));
            this.fIsDirectionLTR = this.fBidi.baseIsLeftToRight();
        }
        else {
            this.fLevels = null;
            this.fIsDirectionLTR = true;
        }
        try {
            this.fComponents = TextLine.getComponents(this.fParagraph, this.fChars, n, n2, inverseMap, this.fLevels, textLabelFactory);
        }
        catch (IllegalArgumentException ex) {
            System.out.println("startingAt=" + n + "; endingAt=" + n2);
            System.out.println("fComponentLimit=" + this.fComponentLimit);
            throw ex;
        }
        this.fComponentStart = n;
        this.fComponentLimit = n2;
    }
    
    private int calcLineBreak(int n, float n2) {
        int fComponentStart = this.fComponentStart;
        int i = 0;
        while (i < this.fComponents.length) {
            int n3 = fComponentStart + this.fComponents[i].getNumCharacters();
            if (n3 > n) {
                break;
            }
            fComponentStart = n3;
            ++i;
        }
        while (i < this.fComponents.length) {
            TextLineComponent textLineComponent = this.fComponents[i];
            int numCharacters = textLineComponent.getNumCharacters();
            int lineBreakIndex = textLineComponent.getLineBreakIndex(n - fComponentStart, n2);
            if (lineBreakIndex == numCharacters && i < this.fComponents.length) {
                n2 -= textLineComponent.getAdvanceBetween(n - fComponentStart, lineBreakIndex);
                fComponentStart = (n = fComponentStart + numCharacters);
                ++i;
            }
            else {
                return fComponentStart + lineBreakIndex;
            }
        }
        if (this.fComponentLimit < this.fChars.length) {
            this.generateComponents(n, this.fChars.length);
            return this.calcLineBreak(n, n2);
        }
        return this.fChars.length;
    }
    
    private int trailingCdWhitespaceStart(int n, int n2) {
        if (this.fLevels != null) {
            int n3 = (byte)!this.fIsDirectionLTR;
            while (--n2 >= n) {
                if (this.fLevels[n2] % 2 != n3 && Character.getDirectionality(this.fChars[n2]) == 12) {
                    continue;
                }
                return ++n2;
            }
        }
        return n;
    }
    
    private TextLineComponent[] makeComponentsOnRange(int i, int b) {
        int trailingCdWhitespaceStart = this.trailingCdWhitespaceStart(i, b);
        int fComponentStart = this.fComponentStart;
        int j = 0;
        while (j < this.fComponents.length) {
            int n = fComponentStart + this.fComponents[j].getNumCharacters();
            if (n > i) {
                break;
            }
            fComponentStart = n;
            ++j;
        }
        boolean n2 = false;
        for (int k = 1; k != 0; ++j) {
            int a = fComponentStart + this.fComponents[j].getNumCharacters();
            if (trailingCdWhitespaceStart > Math.max(fComponentStart, i) && trailingCdWhitespaceStart < Math.min(a, b)) {
                n2 = true;
            }
            if (a >= b) {
                k = 0;
            }
            else {
                fComponentStart = a;
            }
        }
        int n3 = j - j;
        if (!(n2 == 0)) {
            ++n3;
        }
        TextLineComponent[] array = new TextLineComponent[n3];
        int n4 = 0;
        int n5;
        if ((trailingCdWhitespaceStart = trailingCdWhitespaceStart) == (i = i)) {
            n5 = !this.fIsDirectionLTR;
            trailingCdWhitespaceStart = b;
        }
        else {
            n5 = 2;
        }
        while (i < b) {
            int b2 = fComponentStart + this.fComponents[j].getNumCharacters();
            int max = Math.max(i, fComponentStart);
            int min = Math.min(trailingCdWhitespaceStart, b2);
            array[n4++] = this.fComponents[j].getSubset(max - fComponentStart, min - fComponentStart, n5);
            i += min - max;
            if (i == trailingCdWhitespaceStart) {
                trailingCdWhitespaceStart = b;
                n5 = (this.fIsDirectionLTR ? 0 : 1);
            }
            if (i == b2) {
                ++j;
                fComponentStart = b2;
            }
            else {
                continue;
            }
        }
        return array;
    }
    
    private TextLine makeTextLineOnRange(int lineStart, int lineLimit) {
        int[] inverseMap = null;
        byte[] levels = null;
        if (this.fBidi != null) {
            levels = BidiUtils.getLevels(this.fBidi.createLineBidi(lineStart, lineLimit));
            inverseMap = BidiUtils.createInverseMap(BidiUtils.createVisualToLogicalMap(levels));
        }
        return new TextLine(this.fFrc, this.makeComponentsOnRange(lineStart, lineLimit), this.fBaselineOffsets, this.fChars, lineStart, lineLimit, inverseMap, levels, this.fIsDirectionLTR);
    }
    
    private void ensureComponents(int n, int n2) {
        if (n < this.fComponentStart || n2 > this.fComponentLimit) {
            this.generateComponents(n, n2);
        }
    }
    
    private void makeLayoutWindow(int preceding) {
        int offset = this.fChars.length;
        if (this.layoutCount > 0 && !this.haveLayoutWindow) {
            offset = Math.min(preceding + (int)((float)Math.max(this.layoutCharCount / this.layoutCount, 1) * TextMeasurer.EST_LINES), this.fChars.length);
        }
        if (preceding > 0 || offset < this.fChars.length) {
            if (this.charIter == null) {
                this.charIter = new CharArrayIterator(this.fChars);
            }
            else {
                this.charIter.reset(this.fChars);
            }
            if (this.fLineBreak == null) {
                this.fLineBreak = BreakIterator.getLineInstance();
            }
            this.fLineBreak.setText(this.charIter);
            if (preceding > 0 && !this.fLineBreak.isBoundary(preceding)) {
                preceding = this.fLineBreak.preceding(preceding);
            }
            if (offset < this.fChars.length && !this.fLineBreak.isBoundary(offset)) {
                offset = this.fLineBreak.following(offset);
            }
        }
        this.ensureComponents(preceding, offset);
        this.haveLayoutWindow = true;
    }
    
    public int getLineBreakIndex(int n, float n2) {
        int n3 = n - this.fStart;
        if (!this.haveLayoutWindow || n3 < this.fComponentStart || n3 >= this.fComponentLimit) {
            this.makeLayoutWindow(n3);
        }
        return this.calcLineBreak(n3, n2) + this.fStart;
    }
    
    public float getAdvanceBetween(int n, int n2) {
        int n3 = n - this.fStart;
        int n4 = n2 - this.fStart;
        this.ensureComponents(n3, n4);
        return this.makeTextLineOnRange(n3, n4).getMetrics().advance;
    }
    
    public TextLayout getLayout(int n, int n2) {
        int n3 = n - this.fStart;
        int n4 = n2 - this.fStart;
        this.ensureComponents(n3, n4);
        TextLine textLineOnRange = this.makeTextLineOnRange(n3, n4);
        if (n4 < this.fChars.length) {
            this.layoutCharCount += n2 - n;
            ++this.layoutCount;
        }
        return new TextLayout(textLineOnRange, this.fBaseline, this.fBaselineOffsets, this.fJustifyRatio);
    }
    
    private void printStats() {
        System.out.println("formattedChars: " + this.formattedChars);
        this.collectStats = false;
    }
    
    public void insertChar(AttributedCharacterIterator paragraph, int index) {
        if (this.collectStats) {
            this.printStats();
        }
        if (TextMeasurer.wantStats) {
            this.collectStats = true;
        }
        this.fStart = paragraph.getBeginIndex();
        int endIndex = paragraph.getEndIndex();
        if (endIndex - this.fStart != this.fChars.length + 1) {
            this.initAll(paragraph);
        }
        char[] array = new char[endIndex - this.fStart];
        int start = index - this.fStart;
        System.arraycopy(this.fChars, 0, array, 0, start);
        array[start] = paragraph.setIndex(index);
        System.arraycopy(this.fChars, start, array, start + 1, endIndex - index - 1);
        this.fChars = array;
        if (this.fBidi != null || Bidi.requiresBidi(array, start, start + 1) || paragraph.getAttribute(TextAttribute.BIDI_EMBEDDING) != null) {
            this.fBidi = new Bidi(paragraph);
            if (this.fBidi.isLeftToRight()) {
                this.fBidi = null;
            }
        }
        this.fParagraph = StyledParagraph.insertChar(paragraph, this.fChars, index, this.fParagraph);
        this.invalidateComponents();
    }
    
    public void deleteChar(AttributedCharacterIterator paragraph, int n) {
        this.fStart = paragraph.getBeginIndex();
        int endIndex = paragraph.getEndIndex();
        if (endIndex - this.fStart != this.fChars.length - 1) {
            this.initAll(paragraph);
        }
        char[] fChars = new char[endIndex - this.fStart];
        int n2 = n - this.fStart;
        System.arraycopy(this.fChars, 0, fChars, 0, n - this.fStart);
        System.arraycopy(this.fChars, n2 + 1, fChars, n2, endIndex - n);
        this.fChars = fChars;
        if (this.fBidi != null) {
            this.fBidi = new Bidi(paragraph);
            if (this.fBidi.isLeftToRight()) {
                this.fBidi = null;
            }
        }
        this.fParagraph = StyledParagraph.deleteChar(paragraph, this.fChars, n, this.fParagraph);
        this.invalidateComponents();
    }
    
    char[] getChars() {
        return this.fChars;
    }
    
    static {
        TextMeasurer.EST_LINES = 2.1f;
        TextMeasurer.wantStats = false;
    }
}
